// vMac 0.x Source
// By Philip Cummins (PC)
// Copyright PC 1997

// PC - ROM Function Files

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "ROM.h"
#include "prefs.h"

#include "memory.h"

#ifdef _ROM_Debug
#include <stdio.h>
#endif

UWORD private_ROM [kROM_Size/2];
UWORD *ROM = private_ROM;

int  ROM_Init (void)
{
  FILE *ROM_File;
  int  File_Size;
  int i;

  /* Gemulator ROM board support */
  if(thePrefs.CurrentROMPath==255)
 {
      File_Size=Gemulator_LoadROM(ROM, thePrefs.GemulatorBase, thePrefs.GemulatorSocket, 2); //2 for MacPlus ROM set

      if (File_Size < kROM_Size)
      {
        printf("vMac: Wrong Gemulator ROM board setting(s) in vMac.ini.\n");

        vMac_Message("ERROR! Wrong Gemulator ROM board setting(s) in vMac.ini.");

        return 1;
      }
 }
  else
 {
  ROM_File = fopen(thePrefs.ROMPath[thePrefs.CurrentROMPath], "rb");
  if (ROM_File == NULL)
    ROM_File = fopen("vMac.ROM","rb");
  if (ROM_File == NULL)
  {
    vMac_Message("vMac.ROM could not be found.\nRead the ReadMe file before contacting us about this.");
    return 1;
  }
  printf("Loading vMac.ROM...\n");
  File_Size = fread(ROM, kROM_Size, 1, ROM_File);
  if (File_Size != 1)
  {
    vMac_Message("vMac.ROM is not 128 KB.");
    fclose(ROM_File);
    return 1;
  }
  fclose(ROM_File);
 }

  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEEE1))
  {
    printf("vMac.ROM loaded successfully, Mac Plus ROM v 1, 'Lonely Hearts'.");
    return 0;
  }
  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEAE1))
  {
    printf("vMac: vMac.ROM loaded successfully, Mac Plus ROM v 2, 'Lonely Heifers'.");
    return 0;
  }
  if ((do_get_mem_word(&ROM[0]) == 0x4D1F)
   && (do_get_mem_word(&ROM[1]) == 0x8172))
  {
    printf("vMac: vMac.ROM loaded successfully, Mac Plus ROM v 3, 'Loud Harmonicas'.");
    return 0;
  }

  {
    char Temp[256];

    sprintf(Temp,"vMac.ROM loaded successfully, but we don't support this ROM yet.\nChecksum = %X, %X\nRaw = %X, %X",do_get_mem_word(&ROM[0]),do_get_mem_word(&ROM[1]),ROM[0],ROM[1]);

    vMac_Message(Temp);
  }

   return 1;
}

int ROM_GetVer(void)
{
  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEEE1))
    {return(1);}

  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEAE1))
    {return(2);}

  if ((do_get_mem_word(&ROM[0]) == 0x4D1F)
   && (do_get_mem_word(&ROM[1]) == 0x8172))
    {return(3);}

    return(0);
}

void  ROM_Reset (void)
{
  extern UBYTE Sony_Installed;

  if (Sony_Installed == 1)
    (void) ROM_Init();
}

UWORD *ROM_Xlate (CPTR addr)
{
  return ROM + (addr >> 1);
}

UBYTE ROM_GetB (CPTR addr)
{
  return ROM [addr >> 1] >> (addr & 1 ? 0 : 8);
}

UWORD ROM_GetW (CPTR addr)
{
  return ROM [addr >> 1];
}

ULONG ROM_GetL (CPTR addr)
{
  return ((ULONG) ROM [addr >> 1] << 16) | ROM [(addr >> 1) + 1];
}

void  ROM_PutB (CPTR addr, UBYTE in)
{
  return;
}

void  ROM_PutW (CPTR addr, UWORD in)
{
  return;
}

void  ROM_PutL (CPTR addr, ULONG in)
{
  return;
}

