// PC - Sony Driver Functions

// This sony code can read 1.4MB Mac floppies and Mac CD-ROM's under Win32
// Drive 1 is reserved for Mac floppies, and drive 4 is for Mac CD-ROM's

// This sony code can also read disk images in addition to DSK/HFV files

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"

#include "memory.h"
#include "Memory1.h"
#include "custom.h"

#include "readcpu.h"
#include "newcpu.h"

#include "Native.h"
#include "Sony.h"
#include "prefs.h"
#include "Screen.h"

#include "reentrant_newcpu.h"

#include <ctype.h>
#include "debug.h"

#ifdef _Sony_Debug
#include <stdio.h>
#endif

#include "Win32\Win32Port.h"
#include <malloc.h>
#include <io.h>

#include "patch.h"

#define _Sony_Writable

UBYTE Sony_Installed = 0;
UBYTE Sony_Opened = 0;
CPTR Sony_Table[5];
CPTR ParamBlk = 0;
CPTR DeviceCtl = 0;

int   Floppy = NULL; // Drive 1 (1.4 Mac floppy)
FILE *Drive2 = NULL; // Drive 2 (DrivePath1)
FILE *Drive3 = NULL; // Drive 3 (DrivePath2)
int   CD = NULL;     // Drive 4 (CD-ROM)

BOOL Drive2Image=FALSE;
BOOL Drive3Image=FALSE;

unsigned char *dread(ULONG start, ULONG count);
int  dwrite(char *_write_buf, ULONG start, ULONG count);
int  dcheck2(void);
int  dopen(void);
void  dclose(void);

unsigned char *CDread(ULONG start, ULONG count);
int CDcheck();

int Sony_Init (void)
{
  Sony_Installed = 0;
  Sony_Opened = 0;

  return 0;
}

void Sony_Reset (void)
{
  CPTR SonyVars = get_long(0x0134);

  if (Floppy)
  {
      put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);      // Drive 1 Writeable
      put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x00);    // Drive 1 No Disk
      put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0x00);     // Drive 1 Single Format (Initially)
      dclose();
      Floppy=NULL;
  }

  if (CD)
  {
      put_byte(SonyVars + kDrive4 + kWriteProt, 0x00);      // Drive 4 Writeable
      put_byte(SonyVars + kDrive4 + kDiskInPlace, 0x00);    // Drive 4 No Disk
      put_byte(SonyVars + kDrive4 + kTwoSideFmt, 0x00);     // Drive 4 Single Format (Initially)
      CD=NULL;
  }

  if (Drive2)
    Sony_Eject(2);

  if (Drive3)
    Sony_Eject(3);

  (void) Sony_Init();
}

// Check_Sony_Patch
// Checks to see if the Sony driver is being read, then patches it up with Sony_Install

int Check_Sony_Patch (void)
{
    return(patch_ALine());
}

// Check_Sony_Illegal
// Checks to see if the ILLEGAL was a call to our Sony driver
int Check_Sony_Illegal (void)
{


  if (Sony_Table[kOpen] == m68k_getpc())
  {
    Sony_Open();
    return 1;
  }
  if (Sony_Table[kPrime] == m68k_getpc())
  {
    Sony_Prime();
    return 1;
  }
  if (Sony_Table[kControl] == m68k_getpc())
  {
    Sony_Control();
    return 1;
  }
  if (Sony_Table[kStatus] == m68k_getpc())
  {
    Sony_Status();
    return 1;
  }
  if (Sony_Table[kClose] == m68k_getpc())
  {
    Sony_Close();
    return 1;
  }
  return 0;
}

// Sony_Install
// This installs a dummy ILLEGAL+RTS header over the Sony driver
// It allows us to trap ILLEGAL to our own routines
void Sony_Install (void)
{
  CPTR VectorPos = 0, SonyPtr = 0;
  int Counter = 0;

  VectorPos = SonyPtr = get_long(get_long(m68k_areg(regs, 7))) & 0x00FFFFFF;

ROMmem_write_enable_set(on);

	SonyPtr += sizeof(UWORD) * 4; // Skip Driver Header (4 Words)

	put_word(SonyPtr, 0x0018); SonyPtr += sizeof(UWORD); // Adjust Open Offset
	put_word(SonyPtr, 0x001C); SonyPtr += sizeof(UWORD); // Adjust Prime Offset
	put_word(SonyPtr, 0x0020); SonyPtr += sizeof(UWORD); // Adjust Control Offset
	put_word(SonyPtr, 0x0024); SonyPtr += sizeof(UWORD); // Adjust Status Offset
	put_word(SonyPtr, 0x0028); SonyPtr += sizeof(UWORD); // Adjust Close Offset

	SonyPtr += sizeof(UWORD) * 3; // Skip Driver Name (3 Words)

	VectorPos += 0x0018; // Update Vector Position

	for (Counter = 0; Counter <= 4; Counter++)
	{
// Install ILLEGAL instruction
		put_word(SonyPtr, 0x4AFC); SonyPtr += sizeof(UWORD);
// Install RTS instruction
		put_word(SonyPtr, 0x4E75); SonyPtr += sizeof(UWORD);

		Sony_Table[Counter] = VectorPos;
		VectorPos += 4;
	}

ROMmem_write_enable_set(off);

	Sony_Installed = 1;
}

// Sony_Return
// Returns from a Sony driver call properly
// Sets D0 to be the Error code, or leaves it alone if SaveD0 = 1
// If Mode = 1, this is an immediate call for sure, use RTS to exit
// Otherwise check it ourselves and RTS or jIODone
void Sony_Return (int iErr, int SaveD0, int Mode)
{
  put_word(ParamBlk + kioResult, iErr); // Update ioResult
  if (SaveD0 == 0) { // Don't Save D0
    if ((iErr & 0x8000) == 0x8000)
      m68k_dreg(regs, 0) = 0xFFFF0000 | iErr; // Mask D0 properly
    else
      m68k_dreg(regs, 0) = 0x00000000 | iErr; // Clear D0 properly
  }
  // Synchronous, Asynchronous or Immediate
  if (((get_word(ParamBlk + kioTrap) & 0x0200) == 0x0200) || (Mode == 1))
    m68k_setpc(m68k_getpc() + 2); // Immediate Return (RTS)
  else
    m68k_setpc(get_long(0x08FC)); // Jump to JIODone (always done)
}

void Sony_Insert (UWORD Drive_No, char *filename)
{
  CPTR SonyVars = get_long(0x0134);
  ULONG tA0 = m68k_areg(regs, 0);
  ULONG tD0 = m68k_dreg(regs, 0);

  if (Sony_Opened == 0)
    return; // Sony driver isn't open yet, ignore

   if (Drive_No == 0x0001) {
        int result;

          if(Floppy!=NULL)
            return;

            result=dcheck();
            if (result==0x80)
             {Floppy=NULL;
              return;}
            else
             {
               if (result==0x03)
                  put_byte(SonyVars + kDrive1 + kWriteProt, 0xFF);
               if (result==0x00)
                  put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);

               #ifndef _Sony_Writable
                put_byte(SonyVars + kDrive1 + kWriteProt, 0xFF);
               #endif

               Floppy=1;
               dopen();

//               SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Insert: Floppy");
             }
      put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x01);    // Drive 1 Disk Inserted
      put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0xFF);     // Drive 1 Double Format
   }

   if (Drive_No == 0x0002) {
      UBYTE Temp[84];

      if(Drive2!=NULL)
        return;

      #ifdef _Sony_Writable
         Drive2 = fopen(filename, "rb+");
         if (Drive2 == NULL)
            return;                                        // Error, no such disk file
         put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);  // Drive 2 Unlocked
      #else
         Drive2 = fopen(filename, "rb");
         if (Drive2 == NULL)
            return;                                         // Error, no such disk file
         put_byte(SonyVars + kDrive2 + kWriteProt, 0xFF);   // Drive 2 Locked
      #endif
      put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x01);    // Drive 2 Disk Inserted
      put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0xFF);     // Drive 2 Double Format

      fread (Temp, 1, 84, Drive2);
      if (Temp[81]==34 && Temp[82]==1 && Temp[83]==0)
       {Drive2Image=TRUE;}
      else
       {Drive2Image=FALSE;
        fseek(Drive2, 0, SEEK_SET);}

//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Insert: DSK 1");
   }

   if (Drive_No == 0x0003) {
      UBYTE Temp[84];

      if(Drive3!=NULL)
        return;

      #ifdef _Sony_Writable
         Drive3 = fopen(filename, "rb+");
         if (Drive3 == NULL)
            return;                                         // Error, no such disk file
         put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);   // Drive 3 Unlocked
      #else
         Drive3 = fopen(filename, "rb");
         if (Drive3 == NULL)
            return;                                         // Error, no such disk file
         put_byte(SonyVars + kDrive3 + kWriteProt, 0xFF);   // Drive 3 Locked
      #endif
      put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x01);    // Drive 3 Disk Inserted
      put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0xFF);     // Drive 3 Double Format

      fread (Temp, 1, 84, Drive3);
      if (Temp[81]==34 && Temp[82]==1 && Temp[83]==0)
       {Drive3Image=TRUE;}
      else
       {Drive3Image=FALSE;
        fseek(Drive3, 0, SEEK_SET);}

//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Insert: DSK 2");
   }

   if (Drive_No == 0x0004) {
          int result;

          if(CD!=NULL)
            return;

            result=CDcheck();
            if (result==0x80)
             {CD=NULL;
              return;}

      CD=1;
      put_byte(SonyVars + kDrive4 + kWriteProt, 0x00);
      put_byte(SonyVars + kDrive4 + kDiskInPlace, 0x01);    // Drive 4 Disk Inserted
      put_byte(SonyVars + kDrive4 + kTwoSideFmt, 0xFF);     // Drive 4 Double Format

//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Insert: CD");
   }

  Sony_Update();
}

void Sony_Eject (UWORD Drive_No)
{
  CPTR SonyVars = get_long(0x0134);

   if (Drive_No == 0x0001 && Floppy) {
      put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);      // Drive 1 Writeable
      put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x00);    // Drive 1 No Disk
      put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0x00);     // Drive 1 Single Format (Initially)
//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Eject: Floppy");
      dclose();
      Floppy=NULL;
      if (thePrefs.Floppy==1 || (thePrefs.Floppy==2 && !thePrefs.DX))
        vMac_Message("You may remove the disk.");
      return;
   }

   if (Drive_No == 0x0002 && Drive2) {
      put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);      // Drive 2 Writeable
      put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x00);    // Drive 2 No Disk
      put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0x00);     // Drive 2 Single Format (Initially)
      fclose(Drive2);
      Drive2 = NULL;
//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Eject: DSK 1");
      return;
   }

   if (Drive_No == 0x0003 && Drive3) {
      put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);      // Drive 3 Writeable
      put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x00);    // Drive 3 No Disk
      put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0x00);     // Drive 3 Single Format (Initially)
      fclose(Drive3);
      Drive3 = NULL;
//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Eject: DSK 2");
      return;
   }

   if (Drive_No == 0x0004 && CD) {
      put_byte(SonyVars + kDrive4 + kWriteProt, 0x00);      // Drive 4 Writeable
      put_byte(SonyVars + kDrive4 + kDiskInPlace, 0x00);    // Drive 4 No Disk
      put_byte(SonyVars + kDrive4 + kTwoSideFmt, 0x00);     // Drive 4 Single Format (Initially)
//      SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)"Eject: CD");
      CD=NULL;

      if(thePrefs.EjectCD)
        eject_media((CDdrive+'A'), NULL);

      if (thePrefs.CD==1 || (thePrefs.CD==2 && !thePrefs.DX))
        vMac_Message("You may remove the CD.");
      return;
   }
}

void Sony_Access (UWORD Drive_No)
{
  UWORD *Buffer = NULL;
  ULONG Sony_Start = 0;
  ULONG Sony_Count = 0;

  char * dbuf;

  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);

  Buffer = (void*)((ULONG) get_real_address(0) + (get_long(ParamBlk + kioBuffer) & kMemory_Extent));

  switch (get_word(ParamBlk + kioPosMode) & 0x0F)
  {
    case kfsAtMark :
      Sony_Start = get_long(DeviceCtl + kdCtlPosition); break;
    case kfsFromStart :
      Sony_Start = get_long(ParamBlk + kioPosOffset); break;

    case kfsFromLEOF :
      if(Drive_No==1)
          Sony_Start = 1474560 - get_long(ParamBlk + kioPosOffset);
      else
          Sony_Start = 819200 - get_long(ParamBlk + kioPosOffset); // Gotta fix this!
      break;

    case kfsFromMark :
      Sony_Start = get_long(ParamBlk + kioPosOffset) +
                   get_long(DeviceCtl + kdCtlPosition); break;
  }

  Sony_Count = get_long(ParamBlk + kioReqCount);

  if ( (get_word(ParamBlk + kioTrap) & 0xF0FF ) == 0xA002) // Reading Trap
  {

      if (Drive_No == 1) {
         Sony_LEDs(1,0,((Sony_Count >> 16) / 2));
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            for (i = 0; i < Sony_Count; i++, pos++)
            {
               dbuf=dread(Sony_Start+i, 1);
               if(dbuf==NULL)
                return;

               put_byte(pos, dbuf[0]);
            }
        }
      }

      if (Drive_No == 2) {
         Sony_LEDs(2,0,((Sony_Count >> 16) / 2));
         fseek(Drive2, Sony_Start, SEEK_SET);

         if(Drive2Image)
            fseek(Drive2, 84, SEEK_CUR);

         #ifdef FIXEND
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            for (i = 0; i < Sony_Count; i++, pos++)
               put_byte(pos, fgetc(Drive2));
         }
         #else
           fread(Buffer, 1, Sony_Count, Drive2);
        #endif
      }

      if (Drive_No == 3) {
         Sony_LEDs(3,0,((Sony_Count >> 16) / 2));
         fseek(Drive3, Sony_Start, SEEK_SET);

         if(Drive3Image)
            fseek(Drive3, 84, SEEK_CUR);

         #ifdef FIXEND
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            for (i = 0; i < Sony_Count; i++, pos++) {
               put_byte(pos, fgetc(Drive3));
            }
         }
         #else
            fread(Buffer, 1, Sony_Count, Drive3);
         #endif
      }

      if (Drive_No == 4) {
         Sony_LEDs(4,0,((Sony_Count >> 16) / 2));
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            dbuf = CDread(Sony_Start, Sony_Count);
            if(dbuf==NULL)
                return;

            for (i = 0; i < Sony_Count; i++, pos++)
            {
               put_byte(pos, dbuf[i]);
            }
         }
       }
  }

  if ( (get_word(ParamBlk + kioTrap) & 0xF0FF ) == 0xA003) // Writing
  {
      if (Drive_No == 1) {
          Sony_LEDs(1,1,((Sony_Count >> 16) / 2));
          {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            dbuf = alloca(Sony_Count);

            for (i = 0; i < Sony_Count; i++, pos++)
            {
                  dbuf[i]=get_byte(pos);
            }

            dwrite(dbuf, Sony_Start, Sony_Count);
          }
      }

      if (Drive_No == 2) {
         Sony_LEDs(2,1,((Sony_Count >> 16) / 2));
         fseek(Drive2, Sony_Start, SEEK_SET);

         if(Drive2Image)
            fseek(Drive2, 84, SEEK_CUR);

         #ifdef FIXEND
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            for (i = 0; i < Sony_Count; i++, pos++)
                fputc(get_byte(pos), Drive2);
         }
         #else
            fwrite(Buffer, 1, Sony_Count, Drive2);
         #endif
      }

      if (Drive_No == 3) {
         Sony_LEDs(3,1,((Sony_Count >> 16) / 2));
         fseek(Drive3, Sony_Start, SEEK_SET);

         if(Drive3Image)
            fseek(Drive3, 84, SEEK_CUR);

         #ifdef FIXEND
         {
            ULONG pos = get_long(ParamBlk + kioBuffer) & kMemory_Extent;
            int i;

            for (i = 0; i < Sony_Count; i++, pos++)
                fputc(get_byte(pos), Drive3);
         }
         #else
            fwrite(Buffer, 1, Sony_Count, Drive3);
         #endif
      }

      if (Drive_No == 4) {
        //No code here, since we can't write to CD-ROM's
      }
  }

  put_word(ParamBlk + kioResult, 0x0000);
  put_long(ParamBlk + kioActCount, Sony_Count);
  put_long(DeviceCtl + kdCtlPosition, get_long(DeviceCtl + kdCtlPosition) + Sony_Count);
}

int Sony_Drive_Installed (UWORD Drive_No)
{
  if ((Drive_No >= 1 && Drive_No <= 4))
    return 1; // Drive is installed
  else
    return 0; // Drive is not installed
}

int Sony_Disk_Inserted (UWORD Drive_No)
{
     if (Drive_No == 1 && Floppy)
        return 1;

     if (Drive_No == 2 && Drive2)
        return 1;

     if (Drive_No == 3 && Drive3)
        return 1;

     if (Drive_No == 4 && CD)
        return 1;

  return 0;
}

// This checks every VBL to see if the disk has been read
void Sony_Update (void)
{

  CPTR SonyVars = get_long(0x0134);
  ULONG tA0 = m68k_areg(regs, 0);
  ULONG tD0 = m68k_dreg(regs, 0);

  if (Sony_Opened == 1)
  {
    if (get_byte(SonyVars + kDrive1 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000001; // Drive 1

      Call_ATrap(0xA02F); // PostEvent Trap

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }

    if (get_byte(SonyVars + kDrive2 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000002; // Drive 2

      Call_ATrap(0xA02F); // PostEvent Trap

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }

    if (get_byte(SonyVars + kDrive3 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000003; // Drive 3

      Call_ATrap(0xA02F); // PostEvent Trap

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }

    if (get_byte(SonyVars + kDrive4 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000004; // Drive 4

      Call_ATrap(0xA02F); // PostEvent Trap

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }
  }
}

void Sony_LEDs (UWORD Drive_No, UBYTE Mode, UWORD Time)
{
   //Disable this function for now
/*
   char Temp[80];

   if(Mode==0)
    strcpy(Temp,"Read: ");
   else
    strcpy(Temp,"Write: ");

   if(Drive_No==1)
    strcpy(&Temp[strlen(Temp)], "Floppy");

   if(Drive_No==2)
    strcpy(&Temp[strlen(Temp)], "DSK 1");

   if(Drive_No==3)
    strcpy(&Temp[strlen(Temp)], "DSK 2");

   if(Drive_No==4)
    strcpy(&Temp[strlen(Temp)], "CD");

   SendMessage(hwndStatus, SB_SETTEXT, 3, (LPARAM)Temp);
*/
  return;
}

// Sony_Open
//
// This opens up the Sony driver, it allocates a pointer for the variables
// Sets SonyVars, then sets up the drive variables with correct values
// And then adds the drives into the drive queue
//
// Bill Huey comment ---
//
// It wasn't obvious initially, but what "Opening" the ".Sony" driver doesn't
// mean opening a file, which is pretty stupid of me. I keeped wondering
// where and how the drive structures where being set at start up, only to realize
// that it's right under my nose in the function Sony_Open.
// That's what I get for not paying attention...
//
void Sony_Open (void)
{
  CPTR tA0 = m68k_areg(regs, 0);
  CPTR tD0 = m68k_dreg(regs, 0);
  CPTR SonyVars = 0;

  if (Sony_Opened == 1) // Don't Open Again
  {
    Sony_Return(0x0000,0,1); // Immediate Return, No Error
    return;
  }

  Sony_Opened = 1;

  tA0 = m68k_areg(regs, 0);
  tD0 = m68k_dreg(regs, 0);

  m68k_areg(regs, 0) = 0x00000000;
  m68k_dreg(regs, 0) = 0x00000310; // Size of Sony Variables

// bill huey ---
//
// The HANDLE is return in the A0 register...
// The error code is return in the D0 register...
//

  Call_ATrap(0xA71E); // _NewPtr,SYS,CLEAR Trap (bill comment ---)

  put_long(0x0134,m68k_areg(regs, 0)); // Save it to SonyVars
  SonyVars = m68k_areg(regs, 0);

  // Add Drive 1
  put_word(SonyVars + kDrive1 + kTrack, 0x0000);     // Drive 1 at Track 0
  put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);   // Drive 1 Writeable
  put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x00); // Drive 1 No Disk
  put_byte(SonyVars + kDrive1 + kInstalled, 0x01);   // Drive 1 Installed
  put_byte(SonyVars + kDrive1 + kSides, 0xFF);       // Drive 1 Double Sided
  put_word(SonyVars + kDrive1 + kQDriveNo, 0x0001);  // Drive 1 is Drive 1
  put_word(SonyVars + kDrive1 + kQRefNum, 0xFFFB);   // Drive 1 uses .Sony
  put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0xFF);  // Drive 1 Double Format
  put_byte(SonyVars + kDrive1 + kNewIntf, 0xFF);     // Drive 1 uses new interface
  put_word(SonyVars + kDrive1 + kDriveErrs, 0x0000); // Drive 1 has no errors

  m68k_areg(regs, 0) = SonyVars + kDrive1 + kQLink;
  m68k_dreg(regs, 0) = 0x0001FFFB; // Drive 1 0001 (Hi Word), Driver FFFB (Lo Word)

  Call_ATrap(0xA04E);     // Add Drive 1..."AddDrive" Trap (bill comment ---)

  // Add Drive 2
  put_word(SonyVars + kDrive2 + kTrack, 0x0000);     // Drive 2 at Track 0
  put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);   // Drive 2 Writeable
  put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x00); // Drive 2 No Disk
  put_byte(SonyVars + kDrive2 + kInstalled, 0x01);   // Drive 2 Installed
  put_byte(SonyVars + kDrive2 + kSides, 0xFF);       // Drive 2 Double Sided
  put_word(SonyVars + kDrive2 + kQDriveNo, 0x0002);  // Drive 2 is Drive 2
  put_word(SonyVars + kDrive2 + kQRefNum, 0xFFFB);   // Drive 2 uses .Sony
  put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0x00);  // Drive 2 Single Format (Initially)
  put_byte(SonyVars + kDrive2 + kNewIntf, 0xFF);     // Drive 2 uses new interface
  put_word(SonyVars + kDrive2 + kDriveErrs, 0x0000); // Drive 2 has no errors

  m68k_areg(regs, 0) = SonyVars + kDrive2 + kQLink;
  m68k_dreg(regs, 0) = 0x0002FFFB; // Drive 2 0002 (Hi Word), Driver FFFB (Lo Word)

  Call_ATrap(0xA04E);     // Add Drive 2..."AddDrive" Trap (bill comment ---)

   // Add Drive 3
   put_word(SonyVars + kDrive3 + kTrack, 0x0000);     // Drive 3 at Track 0
   put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);   // Drive 3 Writeable
   put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x00); // Drive 3 No Disk
   put_byte(SonyVars + kDrive3 + kInstalled, 0x01);   // Drive 3 Installed
   put_byte(SonyVars + kDrive3 + kSides, 0xFF);       // Drive 3 Double Sided
   put_word(SonyVars + kDrive3 + kQDriveNo, 0x0003);  // Drive 3 is Drive 3
   put_word(SonyVars + kDrive3 + kQRefNum, 0xFFFB);   // Drive 3 uses .Sony
   put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0x00);  // Drive 3 Single Format (Initially)
   put_byte(SonyVars + kDrive3 + kNewIntf, 0xFF);     // Drive 3 uses new interface
   put_word(SonyVars + kDrive3 + kDriveErrs, 0x0000); // Drive 3 has no errors

   m68k_areg(regs, 0) = SonyVars + kDrive3 + kQLink;
   m68k_dreg(regs, 0) = 0x0003FFFB; // Drive 3 0003 (Hi Word), Driver FFFB (Lo Word)

   Call_ATrap(0xA04E);                                // Add Drive 3

   // Add Drive 4
   put_word(SonyVars + kDrive4 + kTrack, 0x0000);     // Drive 4 at Track 0
   put_byte(SonyVars + kDrive4 + kWriteProt, 0x00);   // Drive 4 Writeable
   put_byte(SonyVars + kDrive4 + kDiskInPlace, 0x00); // Drive 4 No Disk
   put_byte(SonyVars + kDrive4 + kInstalled, 0x01);   // Drive 4 Installed
   put_byte(SonyVars + kDrive4 + kSides, 0xFF);       // Drive 4 Double Sided
   put_word(SonyVars + kDrive4 + kQDriveNo, 0x0004);  // Drive 4 is Drive 4
   put_word(SonyVars + kDrive4 + kQRefNum, 0xFFFB);   // Drive 4 uses .Sony
   put_byte(SonyVars + kDrive4 + kTwoSideFmt, 0x00);  // Drive 4 Single Format (Initially)
   put_byte(SonyVars + kDrive4 + kNewIntf, 0xFF);     // Drive 4 uses new interface
   put_word(SonyVars + kDrive4 + kDriveErrs, 0x0000); // Drive 4 has no errors

   m68k_areg(regs, 0) = SonyVars + kDrive4 + kQLink;
   m68k_dreg(regs, 0) = 0x0004FFFB; // Drive 4 0004 (Hi Word), Driver FFFB (Lo Word)

   Call_ATrap(0xA04E);                                // Add Drive 4

  m68k_areg(regs, 0) = tA0; // Restore Parameter Block and Device Control Entry
  m68k_dreg(regs, 0) = tD0;

  Sony_Insert(1, NULL);
  Sony_Insert(2, thePrefs.DrivePath[0]);
  Sony_Insert(3, thePrefs.DrivePath[1]);
  Sony_Insert(4, NULL);

  FloppyReady=TRUE;

  Sony_Return(0x0000,0,1); // Immediate Return, No Error
}

// Sony_Prime
// Handles I/O to disks
void Sony_Prime (void)
{
  CPTR SonyVars = get_long(0x0134);
  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);

  // See if the Drive is Installed
  if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
    return;
  }

  // See if a disk is currently inserted
  if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
    return;
  }

  if (get_word(ParamBlk + kioVRefNum) == 0x0001)
    put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x02); // Clamp Drive 1

  if (get_word(ParamBlk + kioVRefNum) == 0x0002)
    put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x02); // Clamp Drive 2

  if (get_word(ParamBlk + kioVRefNum) == 0x0003)
    put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x02); // Clamp Drive 3

  if (get_word(ParamBlk + kioVRefNum) == 0x0004)
    put_byte(SonyVars + kDrive4 + kDiskInPlace, 0x02); // Clamp Drive 4

  if ((get_word(ParamBlk + kioTrap) & 0xF0FF) == 0xA003) // PBWrite Trap
  {
    #ifdef _Sony_Writable
    Sony_Access(get_word(ParamBlk + kioVRefNum)); // Prepare to write a disk
    Sony_Return(0x0000,0,0);
    #else
    // Don't enable writes
    Sony_Return(0xFFD4,0,0); // Write Protected (-44)
    #endif
  }

  if ((get_word(ParamBlk + kioTrap) & 0xF0FF) == 0xA002) // PBRead Trap
  {
    Sony_Access(get_word(ParamBlk + kioVRefNum)); // Prepare to read a disk
    Sony_Return(0x0000,0,0);
  }
}

// Sony_Control
// Implements control csCodes for the Sony driver
void Sony_Control (void)
{
  UWORD csCode = 0;

  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);

  if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
    return;
  }

  csCode = get_word(ParamBlk + kcsCode);

  switch (csCode)
  {
    case kKillIO : // Immediate Return of -1 (-17 in PowerMacs?)
    {
      Sony_Return(0xFFFF,0,1);
      break;
    }
    case kVerifyDisk :
    {
      if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
      {
        Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
        return;
      }
      Sony_Return(0x0000,0,0); // No Error (0)
      break;
    }
    case kEjectDisk :
    {
      Sony_Eject(get_word(ParamBlk + kioVRefNum));
      Sony_Return(0x0000,0,0); // No error (0)
      break;
    }
    case kFormatDisk :
    {
      if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
      {
        Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
        return;
      }

      Sony_Return(0xFFD4,0,0); // Write Protected (-44)
      return;
      break;
    }
    case kReturnDriveInfo :
    {
      if (get_word(ParamBlk + kioVRefNum) == 0x0001)
      {
        put_long(ParamBlk + kcsParam, 0x00000003); // 800K, Primary Internal Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }
      if (get_word(ParamBlk + kioVRefNum) == 0x0002)
      {
        put_long(ParamBlk + kcsParam, 0x00000903); // 800K, Secondary External Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }
      if (get_word(ParamBlk + kioVRefNum) == 0x0003)
      {
        put_long(ParamBlk + kcsParam, 0x00000903); // 800K, Secondary External Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }

      if (get_word(ParamBlk + kioVRefNum) == 0x0004)
      {
        put_long(ParamBlk + kcsParam, 0x00000903); // 800K, Secondary External Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }

      break;
    }
    default :
    {
      Sony_Return(0xFFEF,0,0); // Driver can't respond to this Control call (-17)
      break;
    }
  }
}

// Sony_Status
// Handles the DriveStatus call
void Sony_Status (void)
{
  CPTR SonyVars = get_long(0x0134);
  CPTR Src = 0; CPTR Dst = 0;
  int Counter = 0;
  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);

  if (get_word(ParamBlk + kcsCode) == kDriveStatus) // Drive Status Call
  {
    if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
    {
      Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0001) // Drive 1 Status
    {
      Src = SonyVars + kDrive1;
      Dst = ParamBlk + kcsParam;
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0002) // Drive 2 Status
    {
      Src = SonyVars + kDrive2;
      Dst = ParamBlk + kcsParam;
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0003) // Drive 3 Status
    {
      Src = SonyVars + kDrive3;
      Dst = ParamBlk + kcsParam;
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0004) // Drive 4 Status
    {
      Src = SonyVars + kDrive4;
      Dst = ParamBlk + kcsParam;
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
  }
  else
    Sony_Return(0xFFEE,0,0); // Can't Respond to Status Call (-18)
}

// Sony_Close
// Returns immediate error
void Sony_Close(void)
{
  Sony_Return(0xFFE8,0,1); // Can't Close Driver (-24)
}

#define DREAD_BUFFER_SIZE       512
#define CDREAD_BUFFER_SIZE      65536

static HANDLE dread_handle = 0;
static int dread_A_or_B = 0; // 1 if "B:"
static char *dbuffer = 0;
static char *CDbuffer = 0;

int dopen(void)
{
    dread_handle = floppy_init( dread_A_or_B, TRUE );

	return( dread_handle != 0 );
}

void dclose(void)
{
    clear_cache_by_drive_index(dread_A_or_B); //Clear cache

    if(dread_handle) {
        floppy_final( dread_A_or_B );
	    dread_handle = 0;
	}
}

void dinit(void)
{
	floppy_module_global_init();
	dbuffer = (char *)VirtualAlloc(
				NULL,
				DREAD_BUFFER_SIZE,
				MEM_RESERVE | MEM_COMMIT,
				PAGE_READWRITE );
}

void dfinal(void)
{
        dclose();
	if(dbuffer) VirtualFree( dbuffer, 0, MEM_RELEASE  );
	floppy_module_global_final();
}

unsigned char *dread(ULONG start, ULONG count)
{
        if(UseFloppy==0)
            return(NULL);

        if(!dread_handle && !dopen()) return(NULL);
        if(floppy_read( dread_handle, dread_A_or_B, start, count, dbuffer ))
	{
		return(dbuffer);
	} else {
		return(NULL);
	}
}

int dwrite(char *_write_buf, ULONG start, ULONG count)
{
        if(!dread_handle && !dopen()) return(NULL);
        if(floppy_write( dread_handle, dread_A_or_B, start, count, _write_buf ))
	{
		return(TRUE);
	} else {
		return(FALSE);
	}
}

int dcheck(void)
{
  //Scan drives A: and B: for 1.4MB Mac floppy disks

  int result=0;

  if(UseFloppy==0 || FloppyBusy)
    return(0x80);

  FloppyBusy=TRUE;

  //Try drive A:
  dread_A_or_B=0;
  result=dcheck2();

  if(result==0x80)
   {
     //Drive A: didn't work, try drive B:
     dread_A_or_B=1;
     result=dcheck2();
     if(result==0x80)
      {//No disks in A: or B:
       dread_A_or_B=0;
       FloppyBusy=FALSE;
       return(result);
      }
   }

  clear_cache_by_drive_index(dread_A_or_B);

  FloppyBusy=FALSE;

  return(result);
}

int dcheck2(void)
{
    int * Temp;

    if(dopen()==0)
     {dclose();
      return(0x80);}

    //Make sure the disk is in the drive
    Temp=dread(0,512);
    if(Temp==NULL)
     {dclose();
      return(0x80);}

    //See if the disk is writable
    if(dwrite(Temp,0,512)==FALSE)
     {dclose();
      return(0x03);}
    else
     {dclose();
      return(0);}
}

CDinit()
{
    OSVERSIONINFO OSVerInfo;

    OSVerInfo.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
    GetVersionEx(&OSVerInfo);

    UseCD=0;

    if(OSVerInfo.dwPlatformId==VER_PLATFORM_WIN32_WINDOWS)
     {VxdPatchOnOff(TRUE);
      UseCD=1;}

    if(OSVerInfo.dwPlatformId==VER_PLATFORM_WIN32_NT)
     {if(CdenableSysInstallStart!=0)
        UseCD=1;
     }

    if(UseCD==1)
     {
        CDbuffer = (char *)VirtualAlloc(NULL,
                CDREAD_BUFFER_SIZE,
				MEM_RESERVE | MEM_COMMIT,
				PAGE_READWRITE );
     }
}

CDfinal()
{
        if(CDbuffer)
            VirtualFree(CDbuffer, 0, MEM_RELEASE);
}

unsigned char *CDread(ULONG start, ULONG count)
{
        if(!UseCD)
            return(NULL);

        if(ReadCD(start+start_address, count, CDbuffer)!=0)
	{
                return(CDbuffer);
	} else {
		return(NULL);
	}
}

int CDcheck()
{
       //Scan drives for a Mac CD-ROM

       char CDrootdir[4];
       char CDletter;

       CDdrive=-1;
       start_address=0;

       if(UseCD==0)
           return(0x80);

       for(CDletter = 'C'; CDletter <= 'Z'; CDletter++)
        {
               wsprintf(CDrootdir, "%c:\\", CDletter);
               if(GetDriveType(CDrootdir) == DRIVE_CDROM)
                {
                       // Found it
                       CDdrive=CDletter-'A';

                       if(NULL!=get_mdb_block_address(NULL, CDdrive, &start_address))
                         return(0); //Success
                }
        }

       //No Mac CD-ROMs found
       CDdrive=-1;
       return(0x80);
}