// Create The Crossword Puzzle Window and Initialize the Structure

#include <windows.h>
#include <windowsx.h>
#include "resource.h"

extern "C"
{
extern HINSTANCE AppInstance;
void RegisterClassSplash();
HWND CreateWindow_Splash(HWND hwndMain);
void SetSplashTime (int st);
}

int SplashTime=2000;

// ----------------------------------------------------------------------------
// Given a window handle, will center the window in the desktop window
//
// PARAMETERS:
//    hWnd  - Window to center in screen
//
VOID CenterWindowInScreen(HWND hWnd)
{
   RECT rcScreen,rcWindow;

   GetWindowRect(GetDesktopWindow(),&rcScreen);
   GetWindowRect(hWnd,&rcWindow);

   SetWindowPos(hWnd,NULL,(rcScreen.right-rcScreen.left-rcWindow.right+rcWindow.left)/2,
      (rcScreen.bottom-rcScreen.top-rcWindow.bottom+rcWindow.top)/2,
      0,0,SWP_NOZORDER|SWP_NOSIZE);
}

void SetSplashTime (int st)
{
  SplashTime=st;
  return;
}

static LRESULT SplashWndProc_OnCreate(HWND hWnd,CREATESTRUCT *lpCreateStruct)
{
   CenterWindowInScreen(hWnd);
   SetTimer(hWnd,0,SplashTime,NULL);
   return 1;
}

static LRESULT SplashWndProc_OnDestroy(HWND hWnd)
{
   KillTimer(hWnd,0);
   return 0;
}

static LRESULT SplashWndProc_OnTimer(HWND hWnd,UINT id)
{
   // Time up?  Ok, go away!
   DestroyWindow(hWnd);
   return 0;
}

static LRESULT SplashWndProc_OnLButtonDown(HWND hWnd,BOOL fDoubleClick,int x,int y,UINT keyFlags)
{
   DestroyWindow(hWnd);
   return 0;
}

static LRESULT SplashWndProc_OnPaint(HWND hWnd)
{
   PAINTSTRUCT ps;
   HDC hdc,hdcMemory;
   HBITMAP hbmpOld,hbmpSplash;
   BITMAP bm;

   hdc=BeginPaint(hWnd,&ps);
   hdcMemory=CreateCompatibleDC(hdc);

   hbmpSplash=LoadBitmap(AppInstance,MAKEINTRESOURCE(IDB_SPLASH));
   GetObject(hbmpSplash,sizeof(BITMAP),&bm);
   hbmpOld=SelectObject(hdcMemory,hbmpSplash);

   BitBlt(hdc,0,0,bm.bmWidth,bm.bmHeight,hdcMemory,0,0,SRCCOPY);
   SelectObject(hdcMemory,hbmpOld);
   DeleteDC(hdcMemory);

   EndPaint(hWnd,&ps);
   DeleteObject(hbmpSplash);

   return 0;
}

LRESULT CALLBACK SplashWndProc(HWND hWnd,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
   switch(uMessage)
   {
      case WM_CREATE:
         return HANDLE_WM_CREATE(hWnd,wParam,lParam,SplashWndProc_OnCreate);
      case WM_DESTROY:
         return HANDLE_WM_DESTROY(hWnd,wParam,lParam,SplashWndProc_OnDestroy);
      case WM_TIMER:
         return HANDLE_WM_TIMER(hWnd,wParam,lParam,SplashWndProc_OnTimer);
      case WM_LBUTTONDOWN:
         return HANDLE_WM_LBUTTONDOWN(hWnd,wParam,lParam,SplashWndProc_OnLButtonDown);

      case WM_PAINT:
         return HANDLE_WM_PAINT(hWnd,wParam,lParam,SplashWndProc_OnPaint);


   }
   return DefWindowProc(hWnd,uMessage,wParam,lParam);
}

void RegisterClassSplash()
{
   WNDCLASS wc;

   wc.style=0;
   wc.lpfnWndProc=(WNDPROC)SplashWndProc;
   wc.cbClsExtra=0;
   wc.cbWndExtra=0;
   wc.hIcon=NULL;
   wc.hInstance=AppInstance;
   wc.hCursor=LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground=GetStockObject(LTGRAY_BRUSH);
   wc.lpszMenuName=NULL;
   wc.lpszClassName="Splash";

   RegisterClass(&wc);
}

HWND CreateWindow_Splash(HWND hwndMain)
{
   HBITMAP hbmpSplash=LoadBitmap(AppInstance,MAKEINTRESOURCE(IDB_SPLASH));
   BITMAP  bm;
   GetObject(hbmpSplash, sizeof(BITMAP), &bm);

   HWND hWnd=CreateWindow("Splash","vMac Win32",WS_VISIBLE|WS_BORDER|WS_POPUP,
                       CW_USEDEFAULT,CW_USEDEFAULT,bm.bmWidth,bm.bmHeight,
                       hwndMain,NULL,AppInstance,NULL);
   SetWindowPos(hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE);

   return hWnd;
}
